/*
 * Subformula printer.
 * takes two files as an input an generates
 * a subformula file runnable on any sat solver
 * 
 * Author: Gabriel Corvera
 * Version: 1.0 (September 6th, 2005)
 */

 #include "printer.h"
 
 struct Formula F; 
 struct Subformula SubF;
 char QBFfilename[255];
 char filename[255];
 char subformula[255];
 int assignment[MAXATOM]; /* stores assignment as 1 or -1, 0 means no value*/
 int existentialCount[MAXATOM]; 
 
 /**
 * Processing file function
 * It reads the input of the QBF formula
 * and fills the data structure.
 */
void processQBFFile(char *name){

  	FILE *fp;
  	int lastc;
  	int nextc;

  	if ((fp = fopen(name,"r"))==NULL) {
    	printf("File '%s' cannot be found.\n",name);
    	exit(1);
  	}

  	while ((lastc =  fgetc(fp)) == 'c'){
  		//read until you get and End of Line (We are consuming the comments of the file)
		while ((nextc = fgetc(fp)) != EOF && nextc != '\n');
	}
    ungetc(lastc,fp);
	/*get the first line of the problem. It must be a CNF formula*/
    if (fscanf(fp,"p cnf %i %i",&F.numatom,&F.numclause) != 2){
		fprintf(stderr,"Bad input file\n");
		exit(-1);
   	}
   	
    if(F.numatom > MAXATOM){
		fprintf(stderr,"ERROR - too many atoms\n");
		exit(-1);
  	}
    if(F.numclause > MAXCLAUSE){
		fprintf(stderr,"ERROR - too many clauses\n");
		exit(-1);
	}

	/* Continue processing file */
	processFile(fp);
}

void processFile(FILE *fp){
  int i,j,lit;
  int *storeptr1,*storeptr2;
  int freestore1=0,freestore2=0;
  int size,sizeU,sizeE;
  int tmp[MAXLENGTH];

  //VARIABLES
  for (i=1;i<=F.numatom;i++){
    	F.atoms[i]=0;
  }
  F.numatomU=0;
  F.numatomE=0;

  //universal variables are read here
  fscanf(fp," a  %i ",&j);
  //printf(" %i ",j);
  while(j!=0){
    F.numatomU++;
    F.atomU[F.numatomU]=j;
    F.atoms[j]=1;
    fscanf(fp,"%i ",&j);
    //printf(" %i ",j);
  }
  printf("\n");
  //existential variables are read here
  fscanf(fp," e  %i ",&j);
  //printf(" %i ",j);
  while(j!=0){
    F.numatomE++;
    F.atomE[F.numatomE]=j;
    F.atoms[j]=2;
    fscanf(fp,"%i ",&j);
    //printf(" %i ",j);
  }
  //printf("\n");

  for(i = 0;i < 2*MAXATOM+1;i++){
    F.numoccurence[i] = 0;
  }


  //FORMULA
  for(i = 0;i < F.numclause;i++){
    	size=0;
    	do{
		if (fscanf(fp,"%i ",&lit) != 1){
	  	fprintf(stderr, "Bad input file\n");
	  	exit(-1);
		}

		//if the literal is not the terminal (0) we fill the temporal Array
		if(lit != 0){
		  tmp[size]=lit;
		  size++;
		  F.numoccurence[lit+MAXATOM]++;		  
		}
	    }while(lit != 0);

    	if(size > MAXLENGTH){
      		printf("ERROR - clause too long\n");
      		exit(-1);
    	}

    sizeU=0;
    sizeE=0;
    for (j=0;j<size;j++){

     if (freestore1 < MAXLENGTH){
       storeptr1 = (int *) malloc( sizeof(int) * STOREBLOCK );
       freestore1 = STOREBLOCK;
       //fprintf(stderr,"allocating memory...\n");
     }
     if (freestore2 < MAXLENGTH){
       storeptr2 = (int *) malloc( sizeof(int) * STOREBLOCK );
       freestore2 = STOREBLOCK;
       //fprintf(stderr,"allocating memory...\n");
     }

     if (F.atoms[abs(tmp[j])]==1){
       if (sizeU==0)
	 F.clauseU[i] = storeptr1;
       sizeU++;
       *(storeptr1++) = tmp[j];
       freestore1--;
     }

     if (F.atoms[abs(tmp[j])]==2){
       if (sizeE==0)
	 F.clauseE[i] = storeptr2;
       sizeE++;
       *(storeptr2++) = tmp[j];
       freestore2--;
     }

    }
    F.sizeU[i]=sizeU;
    F.sizeE[i]=sizeE;

  }
  fclose(fp);	
}

/**
 * Processing Arguments
 * Initializes the filename variables
 * and verifies the correct usage of the
 * program
 */
void parseArgs(char **args,int argc){
	if(argc == 3){
		sprintf(QBFfilename,"%s",args[1]);
		sprintf(filename,"%s",args[2]);
		sprintf(subformula,"subformula.cnf");		
	}else if(argc ==4){
		sprintf(QBFfilename,"%s",args[1]);
		sprintf(filename,"%s",args[2]);
		sprintf(subformula,"%s",args[3]);	
	}
	else{
		printf("\nUsage: %s Formula.cnf Assignation.asg [subformula.cnf]\n",args[0]);
		exit(1);
	}
}

/**
 * Processing file with assignment
 * Process the assignment to the universal quantifiers
 */
void processAssigFile(char *file){
	FILE *fp;
	int t;
	
	if ((fp = fopen(file,"r"))==NULL) {
    	printf("File '%s' cannot be found.\n",file);
    	exit(1);
  	}
   	//universal variables are read here
  	fscanf(fp,"%i ",&t);

	while(t!=0){
		assignment[ABS(t)]=(ABS(t)/t);
	    fscanf(fp,"%i ",&t);	    
	}
		
}

void generateSubformula(){
	int i,j,aux;
	SubF.numclause=0;
	SubF.numatoms=0;
	for(i=0;i<F.numclause;i++){
		SubF.clause[i]=1;
		SubF.numclause++;
		for(j=0;j<F.sizeU[i];j++){
			aux=F.clauseU[i][j];
			if(ABS(aux)*assignment[ABS(aux)]==aux){
				SubF.clause[i]=0;
				SubF.numclause--;
				break;
			}
		}			
		if(SubF.clause[i]==1){
			for(j=0;j<F.sizeE[i];j++){
				aux=F.clauseE[i][j];
				existentialCount[ABS(aux)]++;
				if(existentialCount[ABS(aux)]==1){
					SubF.numatoms++;	
				}
			}
		}			
	}
}

void exitSubformula(){
	FILE *fp;
	int i,j,aux;
	if ((fp = fopen(subformula,"w"))==NULL) {
    	printf("File '%s' cannot be open to write.\n",subformula);
    	exit(1);
  	}		
  	fprintf(fp,"p cnf %i %i\n",SubF.numatoms,SubF.numclause);
  	
	for(i=0;i<F.numclause;i++){	
		if(SubF.clause[i]==1){
			for(j=0;j<F.sizeE[i];j++){
				aux=F.clauseE[i][j];
				fprintf(fp,"%i ",aux);
			}
			fprintf(fp,"0\n");
		}			
	}  	

}
 
int main(int argc,char *args[]){

 	parseArgs(args,argc);
 	processQBFFile(QBFfilename);
 	processAssigFile(filename);
 	generateSubformula();
 	exitSubformula();
 	
 	return 0;
}
